from __future__ import annotations


__all__ = ("EventManager",)


from collections import defaultdict
import typing

from .events.base import BaseEvent  # noqa: TCH001
from .observers.abstract import BaseAbstractObserver


_ObserversT = typing.Dict[str, typing.List[BaseAbstractObserver]]  # noqa: UP006


class EventManager:
    _instance = None
    _subscriptions: typing.ClassVar[_ObserversT] = defaultdict(list)

    def subscribe(self, event_type: str, observer: BaseAbstractObserver) -> None:
        self._subscriptions[event_type].append(observer)

    def unsubscribe(self, event_type: str, observer: BaseAbstractObserver) -> None:
        self._subscriptions[event_type].remove(observer)

    def notify(self, event: BaseEvent) -> None:
        for observer in self._subscriptions[event.event_type]:
            observer.update(event)



from __future__ import annotations

import typing

from django.apps import AppConfig
from django.conf import settings
from django.utils.module_loading import import_string

from .config import default_config
from .event_manager import EventManager
from .observers.abstract import BaseAbstractObserver


class DjangoEventObserverConfig(AppConfig):
    _event_manager: typing.Optional[EventManager] = None  # noqa: UP007
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_event_observer"

    @property
    def app_settings(self):
        result = default_config.DJANGO_EVENT_OBSERVER.copy()
        if getattr(settings, "DJANGO_EVENT_OBSERVER", None) is not None:
            result.update(settings.DJANGO_EVENT_OBSERVER)
        return result

    @property
    def event_manager(self):
        if self._event_manager is None:
            msg = f"{self.__class__.__name__} is not Ready"
            raise ValueError(msg)
        return self._event_manager

    def load_event_manager(self):
        event_manager_class = import_string(self.app_settings["EVENT_MANAGER"])
        if not issubclass(event_manager_class, EventManager):
            msg = f"{event_manager_class} is not a subclass of EventManager"
            raise TypeError(msg)
        if self._event_manager is None:
            self._event_manager = event_manager_class()  # noqa: WPS601

    def ready(self):
        self.load_event_manager()
        self._auto_subscribe_observers()

    def _auto_subscribe_observers(self):
        for auto_observer in self.app_settings.get("AUTO_OBSERVERS", []):
            observer_cls = import_string(auto_observer.get("CLASS"))
            if not issubclass(observer_cls, BaseAbstractObserver):
                msg = f"{observer_cls} is not a subclass of BaseAbstractObserver"
                raise TypeError(msg)
            observer = import_string(auto_observer["CLASS"])(
                *auto_observer.get("INIT_ARGS", []), **auto_observer.get("INIT_KWARGS", {})
            )
            for event_type in auto_observer.get("EVENT_TYPES", []):
                self.event_manager.subscribe(event_type, observer)



from __future__ import annotations


__all__ = ("BaseModelSignalEvent",)

from django.db import models

from ..base import BaseEvent
from .model_event_enums import ModelSignalEventEnum  # noqa: TCH001


class BaseModelSignalEvent(BaseEvent[models.Model]):
    def __init__(self, signal_name: ModelSignalEventEnum, instance: models.Model):
        self.signal_name = signal_name
        self.model_type = type(instance)
        super().__init__("model_signal", instance)

    def __str__(self):
        return f"{self.event_data}:{self.event_type}{self.signal_name}"



__all__ = ("ModelSignalEventEnum",)

# flake8: noqa

try:
    from enum import StrEnum
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):  # type: ignore
        pass


class ModelSignalEventEnum(StrEnum):
    PRE_INIT = "pre_init"
    POST_INIT = "post_init"
    PRE_SAVE = "pre_save"
    POST_SAVE = "post_save"
    PRE_DELETE = "pre_delete"
    POST_DELETE = "pre_delete"
    M2M_CHANGED = "m2m_changed"
    PRE_MIGRATE = "pre_migrate"
    POST_MIGRATE = "post_migrate"



__all__ = ("DJANGO_EVENT_OBSERVER",)

from types import MappingProxyType


DJANGO_EVENT_OBSERVER = MappingProxyType(
    {
        "EVENT_MANAGER": "django_event_observer.event_manager.EventManager",
        "AUTO_OBSERVERS": [],
    }
)



from __future__ import annotations


__all__ = ("BaseEvent",)


from datetime import datetime, timezone
import typing
from uuid import uuid4

from django.apps import apps


if typing.TYPE_CHECKING:
    from ..event_manager import EventManager


_EventDataT = typing.TypeVar("_EventDataT")


class BaseEvent(typing.Generic[_EventDataT]):
    def __init__(self, event_type: str, event_data: _EventDataT):
        self.event_timestamp: float = datetime.now(tz=timezone.utc).timestamp()  # noqa: UP017
        self.event_id: str = str(uuid4())
        self._event_manager: EventManager = apps.get_app_config("django_event_observer").event_manager
        self.event_type = event_type
        self.event_data = event_data

    def __call__(self):
        self.notify()

    def notify(self):
        self._event_manager.notify(self)

    def __str__(self):
        return f"{self.event_data}:{self.event_type}"



__all__ = ("BaseAbstractObserver",)

from abc import ABC, abstractmethod

from ...events.base import BaseEvent


class BaseAbstractObserver(ABC):
    @abstractmethod
    def update(self, event: BaseEvent) -> None: ...
